@@ -7,20 +7,28 @@ Implements the base class for `TextRank` –
 with placeholder methods to be used by the subclasses for algorithm extensions.
 """
 
-from .util import groupby_apply, default_scrubber
-
 from collections import Counter, defaultdict, OrderedDict
 from dataclasses import dataclass
-from icecream import ic  # type: ignore # pylint: disable=E0401,W0611
-from spacy.tokens import Doc, Span, Token  # type: ignore # pylint: disable=E0401
-import graphviz  # type: ignore # pylint: disable=E0401
 import json
 import math
-import networkx as nx  # type: ignore # pylint: disable=E0401
 import pathlib
 import time
 import typing
 
+from icecream import ic  # type: ignore # pylint: disable=E0401,W0611 # lgtm[py/unused-import]
+from spacy.tokens import Doc, Span, Token  # type: ignore # pylint: disable=E0401
+import graphviz  # type: ignore # pylint: disable=E0401
+import networkx as nx  # type: ignore # pylint: disable=E0401
+
+from .util import groupby_apply, default_scrubber
+
+try:
+    import altair as alt  # type: ignore # pylint: disable=E0401
+    import pandas as pd  # type: ignore # pylint: disable=E0401
+except ImportError:
+    _has_altair_and_pandas = False
+else:
+    _has_altair_and_pandas = True
 
 # parameter type annotation for a *stop words* source
 StopWordsLike = typing.Union[ str, pathlib.Path, typing.Dict[str, typing.List[str]] ]
@@ -110,6 +118,17 @@ A data class representing one element in the *unit vector* of the document.
     coord: float
 
 
+@dataclass
+class Paragraph:
+    """
+A data class representing the distance measure for one paragraph.
+    """
+    start: int
+    end: int
+    para_id: int
+    distance: float
+
+
 class BaseTextRankFactory:
     """
 A factory class that provides the document with its instance of
@@ -143,7 +162,7 @@ parts of speech tags to be kept; adjust this if strings representing the POS tag
 the window for neighboring tokens – similar to a *skip gram*
 
     scrubber:
-optional "scrubber" function to clean up punctuation from a token; if `None` then defaults to `pytextrank.default_scrubber`
+optional "scrubber" function to clean up punctuation from a token; if `None` then defaults to `pytextrank.default_scrubber`; when running, PyTextRank will throw a `FutureWarning` warning if the configuration uses a deprecated approach for a scrubber function
 
     stopwords:
 optional dictionary of `lemma: [pos]` items to define the *stop words*, where each item has a key as a lemmatized token and a value as a list of POS tags; may be a file name (string) or a [`pathlib.Path`](https://docs.python.org/3/library/pathlib.html) for a JSON file; otherwise throws a `TypeError` exception
@@ -546,8 +565,23 @@ normalized rank metric
         non_lemma = len([tok for tok in span if tok.pos_ not in self.pos_kept])
         non_lemma_discount = len(span) / (len(span) + (2.0 * non_lemma) + 1.0)
 
-        # use a root mean square (RMS) to normalize the contributions
-        # of all the tokens
+        # NB:
+        # This implements a *point estimate* for the ratio of the span
+        # of an extracted phrase divided by the number of non-lemma
+        # tokens within that span.
+        #
+        # The algorithm defined as in the original paper
+        # [[mihalcea04textrank]](https://derwen.ai/docs/ptr/biblio/#mihalcea04textrank)
+        # only considered multi-word phrases which had adjacent lemmas.
+        # Early use cases in industry (circa 2009) required better recall
+        # of phrases so this approach of using a point estimate here to
+        # "normalize" the rank metric causes the longer phrases to get
+        # discounted when they included too many non-lemma tokens.
+        # In other words, we allow some non-lemmas, but avoid having
+        # the algorithm become too "greedy" as it builds phrases.
+        #
+        # Kudos to @debraj135 who asked for an explanation about this
+        # section of the code.
         phrase_rank = math.sqrt(sum_rank / (len(span) + non_lemma))
 
         return phrase_rank * non_lemma_discount
@@ -562,15 +596,22 @@ Group the phrases by their text content, selecting the span with the
 maximum rank within each group, then collect the ranked phrases into
 an ordered list.
 
+Throws a `FutureWarning` warning if the configuration uses a
+deprecated approach for a scrubber function
+
     all_phrases:
 the raw phrase list
 
     returns:
 an ordered list of ranked phrases
+
         """
-        data: typing.List[typing.Tuple[str, float, Span]] = [
-            (self.scrubber(span.text), rank, span) for span, rank in all_phrases.items()
-        ]
+        try:
+            data: typing.List[typing.Tuple[Span, float, Span]] = [
+                (self.scrubber(span), rank, span) for span, rank in all_phrases.items()
+            ]
+        except AttributeError:
+            raise FutureWarning("Text-based scrubbers are deprecated. Use a `Span` instead.")
 
         keyfunc = lambda x: x[0]
         applyfunc = lambda g: list((rank, spans) for text, rank, spans in g)
@@ -687,12 +728,68 @@ a list of sentence distance measures
         return sent_dist
 
 
+    def segment_paragraphs (
+        self,
+        sent_dist: typing.List[Sentence],
+        ) -> typing.List[Paragraph]:
+        """
+Segment a ranked document into paragraphs.
+
+    sent_dist:
+a list of ranked Sentence data objects
+
+    returns:
+a list of Paragraph data objects
+        """
+        para_elem: typing.List[int] = []
+        para_bounds: typing.List[typing.List[int]] = []
+
+        # first, determine the paragraph boundaries
+        for sent_id, s in enumerate(self.doc.sents):
+            toke_0 = str(s.__getitem__(0))
+            ret_count = sum(map(lambda c: 1 if c == "\n" else 0, toke_0))
+
+            # test for a paragraph boundary
+            if ret_count > 1:
+                if len(para_elem) > 0:
+                    para_bounds.append(para_elem)
+
+                para_elem = []
+
+            # include this sentence
+            para_elem.append(sent_id)
+
+        # then finalize
+        if len(para_elem) > 0:
+            para_bounds.append(para_elem)
+
+        # second, aggregate the distance measures and construct the
+        # Paragraph data objects
+        para_list: typing.List[Paragraph] = []
+
+        for para_id, para_elem in enumerate(para_bounds):
+            sum_dist = [
+                sent_dist[sent_id].distance
+                for sent_id in para_elem
+                ]
+
+            para_list.append(Paragraph(
+                para_id = para_id,
+                start = para_elem[0],
+                end = para_elem[-1],
+                distance = sum(sum_dist) / float(len(sum_dist)),
+                ))
+
+        return para_list
+
+
     def summary (
         self,
         *,
         limit_phrases: int = 10,
         limit_sentences: int = 4,
         preserve_order: bool = False,
+        level: str="sentence",
         ) -> typing.Iterator[str]:
         """
 Run an
@@ -708,30 +805,56 @@ total number of sentences to yield for the extractive summarization
     preserve_order:
 flag to preserve the order of sentences as they originally occurred in the source text; defaults to `False`
 
+    level:
+default extractive summarization with `"sentence"` value; when set as `"paragraph`" get the average score per paragraph then sort the paragraphs to produce the summary
+
     yields:
 texts for sentences, in order
         """
         # build a list of sentence indices sorted by distance
         sent_dist: typing.List[Sentence] = self.calc_sent_dist(limit_phrases)
 
-        top_sent_ids: typing.List[int] = [
-            sent.sent_id
-            for sent in sorted(sent_dist, key=lambda sent: sent.distance)
+        if level == "sentence":
+            top_sent_ids: typing.List[int] = [
+                sent.sent_id
+                for sent in sorted(sent_dist, key=lambda sent: sent.distance)
+                ]
+
+            # truncated to the specified limit
+            limit = min(limit_sentences, len(top_sent_ids))
+            top_sent_ids = top_sent_ids[:limit]
+
+            # optional: sort in ascending order of index to preserve
+            # the order in which sentences appear in the original text
+            if preserve_order:
+                top_sent_ids.sort()
+
+            # extract sentences with the least distance, up to the limit
+            # requested
+            for sent_id in top_sent_ids:
+                yield sent_dist[sent_id].text(self.doc)
+
+        if level == "paragraph":
+            top_sent_ids = [
+                sent_id
+                for p in sorted(self.segment_paragraphs(sent_dist), key=lambda x: x.distance)
+                for sent_id in range(p.start, p.end + 1)
             ]
 
-        # truncated to the specified limit
-        limit = min(limit_sentences, len(top_sent_ids))
-        top_sent_ids = top_sent_ids[:limit]
+            # truncate to the specified limit
+            limit_para = min(limit_sentences, len(top_sent_ids))
+            top_sent_ids = top_sent_ids[:limit_para]
 
-        # optional: sort in ascending order of index to preserve the
-        # order in which sentences appear in the original text
-        if preserve_order:
-            top_sent_ids.sort()
+            # optional: sort in ascending order of index to preserve
+            # the order in which the sentences appear in the original
+            # text
+            if preserve_order:
+                top_sent_ids.sort()
 
-        # extract sentences with the least distance, up to the limit
-        # requested
-        for sent_id in top_sent_ids:
-            yield sent_dist[sent_id].text(self.doc)
+            # extract sentences with the least distance, up to the limit
+            # requested
+            for sent_id in top_sent_ids:
+                yield sent_dist[sent_id].text(self.doc)
 
 
     def write_dot (
@@ -762,3 +885,30 @@ path for the output file; defaults to `"graph.dot"`
 
         with open(path, "w") as f:  # type: ignore
             f.write(dot.source)
+
+
+    def plot_keyphrases (
+        self
+        ) -> typing.Any:
+        """
+Plot a document's keyphrases rank profile using
+[`altair.Chart`](https://altair-viz.github.io/user_guide/generated/toplevel/altair.Chart.html)
+
+Throws an `ImportError` if the `altair` and `pandas` libraries are not installed.
+
+    returns:
+the `altair` chart being rendered
+        """
+        if not _has_altair_and_pandas:
+            raise ImportError("altair and pandas are required to use this method. Install them with `pip install 'pytextrank[viz]'`")
+
+        source = pd.DataFrame([p.__dict__ for p in self.doc._.phrases]).drop("chunks", axis=1).reset_index()
+
+        c = (
+            alt.Chart(source)
+            .mark_bar()
+            .encode(x="index", y="rank", color="count", tooltip=["text", "rank", "count"])
+            .properties(title="Keyphrase profile of the document")
+        )
+
+        return c
